/* 
window.c 
ALICE V 2.0, January 1993
library functions for window handling 
(C) Ralf Kaufmann 1993 
used C compiler: Pure C 
*/ 
 
/* functions contained in this module: 
clip_window 
clear_window 
open_window 
wd_calc_work 
handle_full 
do_redraw 
full_redraw 
handle_window 
get_index_from_handle 
get_free_index 
easy_open_window 
exist_wind 
top_wind 
own_top_wind 
*/ 
 
/* Includefiles */ 
 
# include <ctype.h> 
# include <math.h> 
# include <stdio.h> 
# include <stdlib.h> 
# include <string.h> 
 
# include <aes.h> 
# include <tos.h> 
# include <vdi.h> 
 
 
# include "language.h" 
# include "prvdef.h" 
# include "mewidi.h" 
# include "messages.h" 
# include "scancode.h" 
# include "structs.h" 
# include "extdef.h" 
# include "rsc.h" 
 
 
/*  Declare VDI globals */ 
 
int contrl  [12]; 
int intin   [256],  ptsin     [256]; 
int intout  [256],  ptsout    [256]; 
int work_in [12],   work_out  [57]; 
 
/* Global variables */ 
 
int       ap_id, handle; 
int       bzeich, hzeich, bbox, hbox, gbl_clip[WD_COORD]; 
int       prgquit, prgexit, closeall; 
MFDB      screen; 
MFDB      src, dst;                          /* for raster op's  */ 
OBJECT    *menu_tree; 
OBJECT    *dial_about, *dial_help, *dial_print, *dial_options; 
OBJECT    *dial_search, *dial_replace, *dial_gotoli; 
OBJECT    *dial_info, *dial_selebuf; 
OBJECT    *dial_repl_info, *dial_gotopa; 
OBJECT    *dial_helpfi, *dial_helpms;
OBJECT    *dial_helpbl, *dial_helpwi;
OBJECT    *dial_pinfo;



int       tab_size, tab_size_cfg, backup, backup_cfg;
int       clocky, clocky_cfg, showcr, showcr_cfg;
int       page_length, page_length_cfg, showpos, showpos_cfg;
int       left_margin, left_margin_cfg;
int       header, header_cfg, lines, lines_cfg;
long      line_cfg;
int       line_len_cfg, line_len, eol_warn, eol_warn_cfg;
int       from_page, to_page, page_number, page_offset;
int       sea_direct, sea_direct_cfg;
int       sea_case, sea_case_cfg;
int       sea_start, sea_start_cfg;
int       repl_case, repl_case_cfg;
int       repl_verify, repl_verify_cfg;
int       repl_start, repl_start_cfg;

 
WIND_DATA windows [WD_MAX_WINDOWS];   /* central window structure */ 
BUFFER    buffers [WD_MAX_BUFFERS];   /* buffers for cut/paste    */
BUFFER    include_buffer;             /* buffer for include file  */

char    param_filename  [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    filename_cfg    [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
long    param_line; 
int     prg_drive;
char    prg_path            [FSEL_MAX_PATH_LEN]; 
int     buffer_index;
char    default_name        [FSEL_MAX_FILNAM_LEN] = ""; 
char    save_as_name        [FSEL_MAX_FILNAM_LEN] = ""; 
char    filename            [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    filename_save       [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    default_path        [FSEL_MAX_PATH_LEN]; 
char    save_as_path        [FSEL_MAX_PATH_LEN]; 
char    search_string       [WD_MAX_EDIT_LENGTH]; 
char    search_string_save  [WD_MAX_EDIT_LENGTH]; 
char    text                [MAX_LINE]; 
char    old_name            [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    new_name            [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    info_string         [MAX_LINE];
char    question            [ALERT_MAX_LINES * ALERT_MAX_LINE_LENGTH]; 
char    delete_name         [FSEL_MAX_FILNAM_LEN]; 
char    delete_path         [FSEL_MAX_PATH_LEN]; 
char    delete_filename     [FSEL_MAX_PATH_LEN+FSEL_MAX_FILNAM_LEN]; 
char    gbl_short_line_buf  [MAX_LINE];    /* make these buffers  */
char    gbl_line_buf1       [3*MAX_LINE];  /* global to reduce    */  
char    gbl_line_buf2       [3*MAX_LINE];  /* memory allocation   */  
char    sea_str             [WD_MAX_EDIT_LENGTH]; 
char    sea_str_cfg         [WD_MAX_EDIT_LENGTH]; 
char    repl_str            [WD_MAX_EDIT_LENGTH]; 
char    repl_str_cfg        [WD_MAX_EDIT_LENGTH]; 
char    replby_str          [WD_MAX_EDIT_LENGTH]; 
char    replby_str_cfg      [WD_MAX_EDIT_LENGTH]; 

 
/**********************/ 
/* Start of routines  */ 
/**********************/ 
 
 
 
/* enables/disables clipping the workarea of a window  */ 
 
void  clip_window (w_handle, flag) 
int   w_handle, flag; 
{ 
  /* get some information about the window  */ 
 
  wind_get (w_handle, WF_WORKXYWH, 
            &gbl_clip[0], &gbl_clip[1], &gbl_clip[2], &gbl_clip[3]); 
 
  /* as wind_get () returns width and height we must transform      */ 
  /* the params for vs_clip () into absolute coordinates            */ 
 
  /* x-coordinate for bottom rightmost point: */ 
 
  gbl_clip[2] += gbl_clip[0]; 
  gbl_clip[2]--;                /* disregard the right border line  */ 
 
  /* y-coordinate for bottom rightmost point: */ 
 
  gbl_clip[3] += gbl_clip[1]; 
  gbl_clip[3]--;               /* disregard the bottom border line  */ 
 
  vs_clip (handle, flag, gbl_clip); 
  return; 
}                                       /* end function clip_window */ 
 
 
 
 
/* clears the entire window */ 
 
void  clear_window (w_handle) 
int   w_handle; 
{ 
  int   rect[4]; 
 
 
  wind_get (w_handle, WF_WORKXYWH,            /* get the work area  */ 
            &rect[0], &rect[1], &rect[2], &rect[3]); 
 
  rect[2]--;                    /* disregard the right border line  */ 
  rect[3]--;                   /* disregard the bottom border line  */ 
  v_hide_c (handle);                              /* hide the mouse */ 
  ALLWHITE (rect[0],rect[1],rect[2],rect[3]); 
  v_show_c (handle, 1);                           /* show the mouse */ 
  return; 
}                                     /* end function clear_window  */ 
 
 
 
 
/* powerfull and very general routine for opening a window          */ 
 
int   open_window (w_name, redraw, was, 
                    algn_x, algn_y, algn_w, algn_h, snp, 
                    w_index, doc_w, doc_l) 
char  *w_name;                            /* pointer to namestring  */ 
void  (*redraw)();                    /* pointer to redraw function */ 
int   was,              /* list of elements (sliders, arrows, etc.) */ 
      algn_x, algn_y, 
      algn_w, algn_h, snp,                /* alignment and snapping */ 
      w_index; 
long  doc_w, doc_l;   /* document width  and length (in pixels)     */ 
 
{ 
  int   w_handle, x, y, w, h, maxw_x, maxw_y, maxw_w, maxw_h; 
  int   mx, my, mw, mh, first_mx, first_my, first_mw, first_mh; 
  int   rtc, i, j; 
 
 
  /* get the size of the desktop  */ 
 
  wind_get (DESKTOP, WF_WORKXYWH, &x, &y, &w, &h); 
  wind_calc (WC_WORK, was,                  /* wind_get cannot be   */ 
             x, y, w, h,                    /* invoked now!         */ 
             &maxw_x, &maxw_y, &maxw_w, &maxw_h); 
 
 
  first_mx = maxw_x;                  /* take this as a first guess */ 
  first_my = maxw_y; 
  first_mw = maxw_w; 
  first_mh = maxw_h; 
 
  j = 0;
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  { 
    if (i != w_index && windows[i].handle != -1) j++;
  }
  j = j % WD_WINDOW_TURN;               /* modulo window turn */

  first_mx += j * bzeich;               /* shift it a little bit... */
  first_my += j * hzeich;     
  first_mw -= WD_WINDOW_TURN * bzeich;  /* ...and reduce the size  */
  first_mh -= WD_WINDOW_TURN * hzeich; 

  /* calculate and save the intial settings for the window size     */ 
 
  if (algn_x) first_mx =  align (first_mx + algn_x/2, algn_x); 
  if (algn_h) first_my =  align (first_my + algn_h/2, algn_y); 
  if (algn_w) first_mw =  align (first_mw - algn_w, algn_w) + WD_WIPO; 
  if (algn_h) first_mh =  align (first_mh - algn_h, algn_h) + WD_HEPO; 
 
  /* calculate and save the maximum window size */ 
 
  if (algn_x) maxw_x   =  align (maxw_x + algn_x/2, algn_x); 
  if (algn_h) maxw_y   =  align (maxw_y + algn_h/2, algn_y); 
  if (algn_w) maxw_w   =  align (maxw_w - algn_w, algn_w) + WD_WIPO; 
  if (algn_h) maxw_h   =  align (maxw_h - algn_h, algn_h) + WD_HEPO; 
 
  wind_calc (WC_BORDER, was, 
             maxw_x, maxw_y, maxw_w, maxw_h, 
             &mx, &my, &mw, &mh); 
  
  evnt_timer (WIND_DELAY, 0);     /* wait a little bit... */
  w_handle = wind_create (was, mx, my, mw, mh); 
  if (w_handle < 0)             /* not successfull ?  */ 
  { 
    wind_delete (w_handle); 
    return (w_handle); 
  } 
 
  /* write all important data into window structure */ 
 
  windows[w_index].handle       = w_handle; 
  windows[w_index].work.g_x     = first_mx; 
  windows[w_index].work.g_y     = first_my; 
  windows[w_index].work.g_w     = first_mw; 
  windows[w_index].work.g_h     = first_mh; 
  windows[w_index].max.g_x      = mx; 
  windows[w_index].max.g_y      = my; 
  windows[w_index].max.g_w      = mw; 
  windows[w_index].max.g_h      = mh; 
  windows[w_index].min_width    = windows[w_index].max.g_w; 
  windows[w_index].min_height   = windows[w_index].max.g_h; 
  windows[w_index].elements     = was; 
  windows[w_index].align_x      = algn_x; 
  windows[w_index].align_y      = algn_y; 
  windows[w_index].align_w      = algn_w; 
  windows[w_index].align_h      = algn_h; 
  windows[w_index].snap         = snp; 
  windows[w_index].w_redraw     = redraw; /* ptr to redraw function */ 
  windows[w_index].scroll_x     = bzeich; 
  windows[w_index].scroll_y     = hzeich; 
  windows[w_index].block.set    = FALSE; 
  windows[w_index].block.cs_x   = -1;     /* block description      */ 
  windows[w_index].block.cs_y   = -1; 
  windows[w_index].block.ce_x   = -1; 
  windows[w_index].block.ce_y   = -1; 
  windows[w_index].doc_x        = 0;    /* cf. output function      */ 
  windows[w_index].doc_y        = 0;    /* cf. output function      */ 
  windows[w_index].curs_x       = 1;    /* default is first column  */ 
  windows[w_index].curs_y       = 1;    /* default is first row     */ 
  windows[w_index].curs_xs      = 1; 
  windows[w_index].curs_vis     = FALSE; 
  windows[w_index].doc_width    = doc_w; 
  windows[w_index].doc_length   = doc_l; 
  windows[w_index].hslide_pos   = -1; 
  windows[w_index].vslide_pos   = -1; 
  windows[w_index].hslide_size  = -1; 
  windows[w_index].vslide_size  = -1; 
  memset (windows[w_index].file_name, 
          (int) 0, 
          (size_t) (WD_INFO_LEN-1));
  memset (windows[w_index].wind_name, 
          (int) 0,
          (size_t) (WD_INFO_LEN-1));
  memset (windows[w_index].info_line, 
          (int) 0,
          (size_t) (WD_INFO_LEN-1));
  strcpy (windows[w_index].file_name, w_name); 
  strcpy (windows[w_index].wind_name, w_name); 
  strcpy (windows[w_index].info_line, "");
  windows[w_index].modified     = FALSE; 
  windows[w_index].bg_modified  = FALSE; 
  windows[w_index].full         = FALSE; 
 
  wind_set (w_handle, WF_NAME,              /* set the window name  */ 
            HIGHW(windows[w_index].wind_name), 
            LOWW(windows[w_index].wind_name), 
            0, 0); 

  wind_set (w_handle, WF_INFO,             
            HIGHW(windows[w_index].info_line), 
            LOWW(windows[w_index].info_line), 
            0, 0); 
 
  wind_calc (WC_BORDER, was, 
             first_mx, first_my, first_mw, first_mh, 
             &mx, &my, &mw, &mh); 
 
  graf_growbox (0, 0, 0, 0, mx, my, mw, mh); 
 
  if ((rtc = wind_open (w_handle, mx, my, mw, mh)) == 0) 
  { 
    wind_delete (w_handle); 
    windows[w_index].handle = -1;  /* mark the structure as free    */ 
    return (rtc); 
  } 
  clear_window (w_handle); 
  set_slider_size (w_handle); 
  set_slider_pos  (w_handle); 
  return (w_handle);                    /* return the window handle */ 
}                                       /* end function open_window */ 
 
 
 
 
/* calculates the work area of a window */ 
 
void  wd_calc_work (w_handle, x, y, w, h) 
int   w_handle, x, y, w, h; 
{ 
  int  w_index; 
 
 
  w_index = get_index_from_handle (w_handle); 
  wind_calc (WC_WORK, windows[w_index].elements, x, y, w, h, 
                   &windows[w_index].work.g_x, 
                   &windows[w_index].work.g_y, 
                   &windows[w_index].work.g_w, 
                   &windows[w_index].work.g_h); 
  return; 
}                             /* end function wd_calc_work  */ 
 
 
 
 
/* sets window to its full size/resets it to previous size    */ 
 
int   handle_full (w_handle) 
int   w_handle; 
{ 
  int   w_index; 
  GRECT prev, curr, full; 
 
 
  w_index = get_index_from_handle (w_handle); 
 
  /* get the window coordinates */ 
 
  wind_get (w_handle, WF_CURRXYWH, 
            &curr.g_x, &curr.g_y, &curr.g_w, &curr.g_h); 
 
  wind_get (w_handle, WF_PREVXYWH, 
            &prev.g_x, &prev.g_y, &prev.g_w, &prev.g_h); 
 
  wind_get (w_handle, WF_FULLXYWH, 
            &full.g_x, &full.g_y, &full.g_w, &full.g_h); 
 
  if (rect_equal (&curr, &full)) 
  {                                             /* 1. if start  */ 
    /* window is fully sized, resize to previous size           */ 
 
    graf_shrinkbox (prev.g_x, prev.g_y, prev.g_w, prev.g_h, 
                    full.g_x, full.g_y, full.g_w, full.g_h); 
    wind_set (w_handle, WF_CURRXYWH, 
              prev.g_x, prev.g_y, prev.g_w, prev.g_h); 
    windows[w_index].full = FALSE; 
    wd_calc_work (w_handle, prev.g_x, prev.g_y, 
                  prev.g_w, prev.g_h); 
 
  } 
  else 
  { 
    /* window is to be fully sized  */ 
 
    graf_growbox (curr.g_x, curr.g_y, curr.g_w, curr.g_h, 
                  full.g_x, full.g_y, full.g_w, full.g_h); 
    wind_set (w_handle, WF_CURRXYWH, 
              full.g_x, full.g_y, full.g_w, full.g_h); 
    windows[w_index].full = TRUE; 
    wd_calc_work (w_handle, full.g_x, full.g_y, 
                  full.g_w, full.g_h); 
  }                                               /* 1. if end  */ 
  return (TRUE); 
}                               /* end function handle_full     */ 
 
 
 
 
/* perform a window redraw within the specified area              */ 
 
int   do_redraw (w_handle, x, y, w, h, clear) 
int   w_handle, x, y, w, h, clear; 
{ 
  int   w_index; 
  GRECT list_element, work, intersect; 
 
 
  wind_update (BEG_UPDATE);       /* message to AES   */ 
  w_index = get_index_from_handle (w_handle); 
 
  v_hide_c (handle);              /* hide the mouse   */ 
  cursor (w_handle, 0);           /* hide the cursor  */ 
  wind_get (w_handle, WF_FIRSTXYWH, 
            &list_element.g_x, &list_element.g_y, 
            &list_element.g_w, &list_element.g_h); 
  while (list_element.g_w > 0 && list_element.g_h > 0) 
  {                               /* 1. while start */ 
    work.g_x = x; 
    work.g_y = y; 
    work.g_w = w; 
    work.g_h = h; 
 
    /* do the rectangles intersect ?  */ 
 
    if (rect_intersect (&list_element, &work, &intersect)) 
    {                                               /* 1. if start  */ 
      /* intersect now contains the coordinates of the intersection */ 
 
      gbl_clip[0] = intersect.g_x; 
      gbl_clip[1] = intersect.g_y; 
      gbl_clip[2] = intersect.g_x + intersect.g_w; 
      gbl_clip[2]--;            /* disregard the right border line  */ 
      gbl_clip[3] = intersect.g_y + intersect.g_h; 
      gbl_clip[3]--;            /* disregard the bottom border line */ 
 
      vs_clip (handle, TRUE, gbl_clip); 
      intersect.g_w--; 
      intersect.g_h--; 
      (*windows[w_index].w_redraw) (w_handle, &intersect, clear); 
    }                                                 /* 1. if end  */ 
 
    /* get the coordinates of the next rectangle                    */ 
 
    wind_get (w_handle, WF_NEXTXYWH, 
              &list_element.g_x, &list_element.g_y, 
              &list_element.g_w, &list_element.g_h); 
 
  }                                                 /* 1. while end */ 
 
  vs_clip (handle, FALSE, gbl_clip);    /* switch off the clipping  */ 
  if (w_handle == top_wind ()) cursor (w_handle, 1); 
  v_show_c (handle, TRUE);                        /* show the mouse */ 
  wind_update (END_UPDATE);                       /* message to AES */ 
  return (TRUE); 
}                                         /* end function do_redraw */ 
 
 
 
 
/* performs a redraw of the entire window */ 
 
int   full_redraw (w_handle, clear) 
int   w_handle, clear; 
{ 
  int         w_index, i; 
  LINE_DESC   *p, *next; 
 
 
  wind_update (BEG_UPDATE);
  w_index = get_index_from_handle (w_handle); 
 
  if (windows[w_index].modified) 
  { 
    if (windows[w_index].wind_name[0] != '*') 
    { 
      for (i = (int) strlen (windows[w_index].wind_name) + 1; 
            i > 0; i--) 
      { 
        (windows[w_index].wind_name)[i] = 
          (windows[w_index].wind_name)[i-1]; 
      } 
      (windows[w_index].wind_name)[0] = '*'; 
                                         /* mark text as modified  */ 
      wind_set (w_handle, WF_NAME,              /* set window name */ 
        HIGHW(windows[w_index].wind_name), 
        LOWW(windows[w_index].wind_name), 
        0, 0); 
    } 
  } 
  windows[w_index].doc_width  = 0; /* recalc doc width and length  */ 
  windows[w_index].doc_length = 0; /* before setting the sliders   */ 
  for (p = windows[w_index].head; p != NULL; p = next) 
  { 
    next = p->next; 
    if (p->len * bzeich > windows[w_index].doc_width) 
    { 
      windows[w_index].doc_width = p->len * bzeich; 
    } 
    windows[w_index].doc_length += hzeich; 
  } 
  set_slider_size (w_handle); 
  set_slider_pos (w_handle); 
 
  do_redraw (w_handle,                    /* perform a full redraw  */
             windows[w_index].work.g_x,   /* of the work area       */
             windows[w_index].work.g_y, 
             windows[w_index].work.g_w, 
             windows[w_index].work.g_h, 
             clear); 
  MOUSE_IS_ARROW; 
  wind_update (END_UPDATE);
  return (TRUE); 
}                                       /* end function full_redraw */ 
 
 
 
 
/* window manager   */ 
 
int   handle_window (buffer) 
int   buffer[MSG_BUFFER]; 
{ 
  int         i, w_handle, action, length, pos, active_window; 
  int         w_index, send_buffer[MSG_BUFFER], dummy[4]; 
  long        save, start_y, help; 
  GRECT       curr, full, tmp; 
  LINE_DESC   *p, *next; 
 
 
  wind_update (BEG_UPDATE);
  for (i = 0; i < MSG_BUFFER; i++) send_buffer[i] = 0; 
  w_index = get_index_from_handle (buffer[3]); 
  v_hide_c (handle);                              /* hide the mouse */ 
 
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  {                                             /* hide all cursors */ 
    w_handle = windows[i].handle; 
    if (w_handle != -1) cursor (w_handle, 0); 
  } 
 
  w_handle = buffer[3];
  wind_get (w_handle, WF_TOP, &w_handle, 
            &dummy[0], &dummy[1], &dummy[2]); 
  active_window = (w_handle == buffer[3]);  
                                /* is this the top/active window ?  */
  
  switch (buffer[0]) 
  {                                             /* 1. switch start  */ 
 
    case  WM_REDRAW   :    /* is sent by AES after call to wind_set */ 
                           /* for new window size (cf. WM_FULLED)   */ 
                        
                        set_slider_size (buffer[3]); 
                        set_slider_pos  (buffer[3]); 
                        do_redraw (buffer[3], buffer[4], buffer[5], 
                                     buffer[6], buffer[7], TRUE); 
                        break; 
 
    case  WM_TOPPED   :
    case  WM_NEWTOP   : 
                        wind_set (buffer[3], WF_TOP, 
                                  buffer[3], 0, 0, 0); 
                        w_index = get_index_from_handle (buffer[3]); 
                        cap_check (w_index); 
                        w_handle = buffer[3];
                        start_y = windows[w_index].doc_y / hzeich; 
                        help  = (windows[w_index].curs_y - 1 - 
                                start_y) * hzeich; 
                        help  = max ((long) help, (long) 0); 
                                    /* redraw the cursor position */
                        if (cursor_visible (w_index))
                        {
                          do_redraw (   
                            (int) w_handle, 
                            (int) (windows[w_index].work.g_x +
                                   (windows[w_index].curs_x - 1) * 
                                    bzeich), 
                            (int) (windows[w_index].work.g_y + help), 
                            (int) bzeich, 
                            (int) hzeich, 
                            (int) TRUE
                                    );
                          windows[w_index].bg_modified = FALSE;
                        }
                        cursor (w_handle, 1);
                        break; 
 
    case  WM_CLOSED   : 
                        menu_tnormal (menu_tree, TFILE, FALSE); 
 
                        if (w_index == -1) break; /* only deal with */ 
                                                  /* my own stuff   */ 
                        action = 0; 
                        if (windows[w_index].modified) 
                        { 
                          v_show_c (handle, TRUE); 
# ifdef english
                          strcpy (question, "[2][ Save | ");
# endif 
# ifdef german
                          strcpy (question, "[2][ Datei | "); 
# endif
# ifdef french
                          strcpy (question, "[2][ Sauver | ");
# endif 

                          length = (int) strlen (question); 
                          if (strlen (windows[w_index].file_name) 
                              <= ALERT_MAX_LINE_LENGTH) 
                          { 
                            strcpy (question + length, 
                                      windows[w_index].file_name); 
                          } 
                          else 
                            /* file name only if path is too long */ 
                          { 
                            pos = rk_strrchr 
                                    (windows[w_index].file_name, 
                                     '\\'); 
                            strcpy (question + length, 
                                       windows[w_index].file_name + 
                                        pos); 
                          } 
                          length = (int) strlen (question); 
 
# ifdef english
                          strcpy (question + length, 
                              " | before closing ? ][ Yes | No ]"); 
# endif
# ifdef german
                          strcpy (question + length, 
                              " | speichern ? ][ Ja | Nein ]"); 
# endif
# ifdef french
                          strcpy (question + length, 
                              " | avant de fermer ? ][ Oui | Non ]"); 
# endif

                          action = form_alert (1, question); 
 
                        } 
                        if (action == 1)        /* perform a backup */ 
                        { 
                          MOUSE_IS_BUSY; 
                          if (!handle_menu (TFILE, IBACK))
                          {
                            strcpy (param_filename, "");
                            MOUSE_IS_ARROW;
                            if (prgquit)
                            {
                              put_message (PRG_NOT_TERMINATED); 
                              prgquit = FALSE;
                            }
                            closeall = FALSE;
                            break;                          
                          } 
                          MOUSE_IS_ARROW; 
                        } 
 
                        /* get the "old" coordinates  */ 
 
                        wind_get (buffer[3], WF_CURRXYWH, 
                                  &tmp.g_x, &tmp.g_y, 
                                  &tmp.g_w, &tmp.g_h); 
 
                        /* display a closing box  */ 
 
                        graf_shrinkbox ( 
                                windows[w_index].max.g_x, 
                                windows[w_index].max.g_y, 
                                windows[w_index].min_width / 2, 
                                windows[w_index].min_height / 2, 
                                tmp.g_x, 
                                tmp.g_y, 
                                tmp.g_w, 
                                tmp.g_h); 
 
                        wind_close  (buffer[3]); 
                        wind_delete (buffer[3]); 
 
                        if (windows[w_index].head != NULL) 
                        {                           /* 1. if start  */ 
                          for (p = windows[w_index].head; 
                                p; p = next) 
                          {                         /* 1. for start */ 
                            next = p->next; 
                                          /* free allocated memory  */ 
                            free ((char *) p->line); 
                            free ((LINE_DESC *) p); 
                          }                           /* 1. for end */ 
                        }                             /* 1. if end  */ 
                        windows[w_index].head   = NULL; 
                        windows[w_index].handle = -1; 

                        if (!exist_wind()) closeall = FALSE;
                        if (prgquit || closeall) 
                        {                           /* 1. if start  */ 
                          if (exist_wind ()) 
                                        /* any other open window ?  */ 
                          {                         /* 2. if start  */ 
                          /* send a message to close the top window */ 
 
                          send_buffer[0] = WM_CLOSED; 
                          send_buffer[1] = ap_id; 
                          send_buffer[2] = 0; 
                          send_buffer[3] = top_wind (); 
                          appl_write ((int) ap_id, 
                                      (int) sizeof (send_buffer), 
                                      send_buffer); 
                          } 
                          else 
                          { 
                            if (prgquit)
                            {
                              prgexit = TRUE;  /* all work was done  */
                            } 
                          }                           /* 2. if end  */ 
                        } 
                        else   
                        { 
                          if (!exist_wind ()) 
                          { 
                            menu_ienable (menu_tree, ICLOSE,   FALSE); 
                            menu_ienable (menu_tree, IRDFILFI, FALSE); 
                            menu_ienable (menu_tree, ICLOSALL, FALSE); 
                            menu_ienable (menu_tree, IDOCINFO, FALSE); 
                            menu_ienable (menu_tree, IGOTOLI,  FALSE); 
                            menu_ienable (menu_tree, IGOTOPA,  FALSE); 
                            menu_ienable (menu_tree, ISAVE,    FALSE); 
                            menu_ienable (menu_tree, ISAVEAS,  FALSE); 
                            menu_ienable (menu_tree, IBACK,    FALSE); 
                            menu_ienable (menu_tree, IBACKAS,  FALSE); 
                            menu_ienable (menu_tree, IFIPRINT, FALSE); 
                            menu_ienable (menu_tree, ISEARCH,  FALSE); 
                            menu_ienable (menu_tree, IREPL,    FALSE); 
                            menu_ienable (menu_tree, IREPSEAR, FALSE); 
                            menu_ienable (menu_tree, IBLSTART, FALSE); 
                            menu_ienable (menu_tree, IBLEND,   FALSE); 
                            cap_enable (FALSE); 
                            menu_ienable (menu_tree, IBLPASTE, FALSE); 
                          } 
                          else 
                          { 
                            w_index = 
                             get_index_from_handle (top_wind()); 
                            if (w_index != -1 && own_top_wind ()) 
                            { 
                              cap_check (w_index); 
                            } 
                          } 
                          menu_ienable (menu_tree, IOPEN, TRUE); 
                        }                             /* 1. if end  */ 
                        menu_tnormal (menu_tree, TFILE, TRUE); 
                        break; 
 
    case  WM_FULLED   : 
                        handle_full (buffer[3]); 
 
                        /* AES will now send a WM_REDRAW message    */ 
 
                        set_slider_size (buffer[3]); 
                        set_slider_pos  (buffer[3]);
                        windows[w_index].bg_modified = !active_window;
                        break; 
 
    case  WM_ARROWED  : 
                        scroll_wind (buffer[3], buffer[4]); 
                        windows[w_index].bg_modified = !active_window;
                        break; 
 
    case  WM_HSLID    : 
                        save = windows[w_index].doc_x; 
                        wind_hslide (buffer[3], buffer[4]); 
                        if (windows[w_index].doc_x != save) 
                        { 
                          w_handle = buffer[3];
                           /* send myself a full redraw message  */ 
                          send_buffer[0] = WM_REDRAW; 
                          send_buffer[1] = ap_id; 
                          send_buffer[2] = 0; 
                          send_buffer[3] = buffer[3]; 
                          send_buffer[4] = 
                          windows[w_index].work.g_x; 
                          send_buffer[5] = 
                           windows[w_index].work.g_y; 
                          send_buffer[6] = 
                           windows[w_index].work.g_w; 
                          send_buffer[7] = 
                           windows[w_index].work.g_h; 
                          appl_write ((int) ap_id, 
                                    (int) sizeof (send_buffer), 
                                    send_buffer);
                        } 
                        windows[w_index].bg_modified = !active_window;
                        break; 
 
    case  WM_VSLID    : 
                        save = windows[w_index].doc_y; 
                        wind_vslide (buffer[3], buffer[4]); 
                        if (windows[w_index].doc_y != save) 
                        { 
                          w_handle = buffer[3];
                           /* send myself a full redraw message  */ 
                          send_buffer[0] = WM_REDRAW; 
                          send_buffer[1] = ap_id; 
                          send_buffer[2] = 0; 
                          send_buffer[3] = buffer[3]; 
                          send_buffer[4] = 
                          windows[w_index].work.g_x; 
                          send_buffer[5] = 
                           windows[w_index].work.g_y; 
                          send_buffer[6] = 
                           windows[w_index].work.g_w; 
                          send_buffer[7] = 
                           windows[w_index].work.g_h; 
                          appl_write ((int) ap_id, 
                                    (int) sizeof (send_buffer), 
                                    send_buffer); 
                        } 
                        windows[w_index].bg_modified = !active_window;
                        break; 
 
    case  WM_SIZED    : 
    case  WM_MOVED    : 
                        /* store the current minimum  */ 
 
                        windows[w_index].min_width = 
                         (int) min ( 
                              (long) windows[w_index].min_width, 
                              (long) buffer[6]); 
                        windows[w_index].min_height = 
                         (int) min ( 
                              (long) windows[w_index].min_height, 
                              (long) buffer[7]); 
 
                        /* snap if required, i.e window is beyond   */ 
                        /* its boundary                             */ 
 
                        if (windows[w_index].snap) 
                        { 
                          tmp.g_x = buffer[4]; 
                          tmp.g_y = buffer[5]; 
                          tmp.g_w = buffer[6]; 
                          tmp.g_h = buffer[7]; 
 
                          snap (&windows[w_index].max, &tmp); 
 
                          buffer[4] = tmp.g_x; 
                          buffer[5] = tmp.g_y; 
                          buffer[6] = tmp.g_w; 
                          buffer[7] = tmp.g_h; 
                        } 
 
                        /* get the "old" coordinates  */ 
 
                        wind_get (buffer[3], WF_CURRXYWH, 
                                      &tmp.g_x, &tmp.g_y, 
                                      &tmp.g_w, &tmp.g_h); 
 
                        /* align if required  */ 
 
                        wind_calc (WC_WORK, 
                                windows[w_index].elements, 
                                buffer[4], buffer[5], 
                                buffer[6], buffer[7], 
                                &windows[w_index].work.g_x, 
                                &windows[w_index].work.g_y, 
                                &windows[w_index].work.g_w, 
                                &windows[w_index].work.g_h); 
 
                        if (windows[w_index].align_x) 
                        { 
                          windows[w_index].work.g_x = 
                            align (windows[w_index].work.g_x, 
                              windows[w_index].align_x); 
                          buffer[4] = windows[w_index].work.g_x; 
                        } 
                        if (windows[w_index].align_y) 
                        { 
                          windows[w_index].work.g_y = 
                            align (windows[w_index].work.g_y, 
                              windows[w_index].align_y); 
                          buffer[5] = windows[w_index].work.g_y; 
                        } 
                        if (windows[w_index].align_w) 
                        { 
                          windows[w_index].work.g_w = 
                            align (windows[w_index].work.g_w, 
                              windows[w_index].align_w) + WD_WIPO; 
                          buffer[6] = windows[w_index].work.g_w; 
                        } 
                        if (windows[w_index].align_h) 
                        { 
                          windows[w_index].work.g_h = 
                            align (windows[w_index].work.g_h, 
                               windows[w_index].align_h) + WD_HEPO; 
                          buffer[7] = windows[w_index].work.g_h; 
                        } 
                        wind_calc (WC_BORDER, 
                                windows[w_index].elements, 
                                windows[w_index].work.g_x, 
                                windows[w_index].work.g_y, 
                                windows[w_index].work.g_w, 
                                windows[w_index].work.g_h, 
                                &buffer[4], &buffer[5], 
                                &buffer[6], &buffer[7]); 
 
                        /* display a moving box */ 
 
                        graf_movebox (buffer[6], buffer[7], 
                                      tmp.g_x, tmp.g_y, 
                                      buffer[4], buffer[5]); 
 
                        /* reposition the window with the   */ 
                        /* new coordinates                  */ 
 
                        wind_set (buffer[3], WF_CURRXYWH, 
                                  buffer[4], buffer[5], 
                                  buffer[6], buffer[7]); 
                        set_slider_size (buffer[3]); 
                        set_slider_pos  (buffer[3]); 
 
                        /* is the window fully sized now ?  */ 
 
                        wind_get (buffer[3], WF_CURRXYWH, 
                                  &curr.g_x, &curr.g_y, 
                                  &curr.g_w, &curr.g_h); 
 
                        wind_get (buffer[3], WF_FULLXYWH, 
                                  &full.g_x, &full.g_y, 
                                  &full.g_w, &full.g_h); 
 
                        windows[w_index].full = 
                          rect_equal (&curr, &full); 
                        windows[w_index].bg_modified = !active_window;
                        break; 
 
    default           : 
                        break; 
 
  }                                               /* 1. switch end  */ 
 
  if (top_wind() != -1) cursor (top_wind(), 1); 
                                              /* show "top" cursor  */ 
  v_show_c (handle, TRUE); 
  display_time (clocky);  
  wind_update (END_UPDATE); 
  return (TRUE); 
}                                     /* end function handle_window */ 
 
 
 
 
/* routine for simplified window open   */ 
 
int   easy_open_window (w_name, redraw, w_index) 
char  *w_name;          /* pointer to namestring                  */ 
void  (*redraw)();      /* pointer to function returning int      */ 
int   w_index;          /* index into global structure "windows"  */ 
{ 
  int   w_handle; 
 
 
  w_handle =  open_window (w_name, redraw, 
                          NAME    +     /* what is in a window ?  */ 
                          CLOSER  + 
                          FULLER  + 
                          MOVER   + 
                          INFO    +
                          SIZER   + 
                          UPARROW + 
                          DNARROW + 
                          VSLIDE  + 
                          LFARROW + 
                          RTARROW + 
                          HSLIDE, 
                          bzeich, hzeich, bzeich, hzeich, TRUE, 
                          w_index, 
                          (long) 
                           (windows[w_index].doc_width * bzeich), 
                          (long) 
                           (windows[w_index].doc_length * hzeich)); 
  return (w_handle);                   /* return the window handle */ 
}                                /* end function easy_open_window  */ 
 
 
 
 
/* given the window handle this function returns the w_index       */ 
/* (if any) in the global datastructure "windows", -1 otherwise    */ 
 
int   get_index_from_handle (w_handle) 
int   w_handle; 
{ 
  int  i; 
 
 
  if (w_handle == -1) return (-1); 
 
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  { 
    if (windows[i].handle == w_handle) return (i); 
  } 
  return (-1); 
}                           /* end function get_index_from_handle */ 
 
 
 
 
/* returns the next free index in the global data structure       */ 
/* "windows" (if any), -1 otherwise                               */ 
 
int   get_free_index () 
{ 
  int  i; 
 
 
  if (!more_windows ()) return (-1);    /* will the OS create     */
                                        /* more windows ?         */
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  {                   /* is my own data structure large enough ?  */
  
    if (windows[i].handle == -1) return (i);
                                        /* return the free index  */
  } 
  return (-1);                /* no more window can be created... */
}                           /* end function get_free_index  */ 
 
 

int   more_windows ()       /* checks for the availibility of   */
                            /* one more window                  */
{
  int   extra_handle_1, extra_handle_2; 
  int   x, y, w, h, was;
  
  
  was = NAME    +     /* what is in a window ?  */ 
        CLOSER  + 
        FULLER  + 
        MOVER   + 
        INFO    +
        SIZER   + 
        UPARROW + 
        DNARROW + 
        VSLIDE  + 
        LFARROW + 
        RTARROW + 
        HSLIDE;
                          
  wind_get (DESKTOP, WF_WORKXYWH, &x, &y, &w, &h); 
  evnt_timer (WIND_DELAY, 0);     /* wait a little bit... */
  if ((extra_handle_1 = wind_create (was, x, y, w, h)) < 0) 
  {
    return (FALSE);
  }
  evnt_timer (WIND_DELAY, 0);     /* wait a little bit... */
  if ((extra_handle_2 = wind_create (was, x, y, w, h)) < 0) 
  {
    wind_delete (extra_handle_1);
    return (FALSE);
  }
  evnt_timer (WIND_DELAY, 0);     /* wait a little bit... */
  wind_delete (extra_handle_1);
  wind_delete (extra_handle_2);
  evnt_timer (WIND_DELAY, 0);     /* wait a little bit... */
  return (TRUE);
}


 
int   exist_wind ()  /* returns TRUE if there is any open window,  */ 
                     /* FALSE otherwise                            */ 
{ 
  int   i; 
 
 
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  { 
    if (windows[i].handle != -1)  return (TRUE); 
  } 
  return (FALSE); 
} 
 
 
 
int   top_wind()  /* returns the handle of the top/active window  */ 
                  /* (if it's my own one) or (if it's not my own  */ 
                  /* one) the handle of any of my open windows    */ 
                  /* (if any), -1 otherwise                       */ 
{ 
  int   w_handle, i, dummy[4]; 
 
 
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  { 
    if ((w_handle = windows[i].handle) != -1) 
    { 
      wind_get (w_handle, WF_TOP, &w_handle, 
                &dummy[0], &dummy[1], &dummy[2]); 
      if (get_index_from_handle (w_handle) != -1) 
      { 
        return (w_handle);                  /* it's my own window */ 
      } 
      else                        /* if the window is not mine... */ 
      { 
        return (windows[i].handle); 
      } 
    } 
  } 
  return (-1); 
} 
 
 
 
int   own_top_wind()        /* returns TRUE if the top window is  */ 
                            /* one of mine, FALSE otherwise       */ 
{ 
  int   w_handle, i, dummy[4]; 
 
 
  for (i = 0; i < WD_MAX_WINDOWS; i++) 
  { 
    if ((w_handle = windows[i].handle) != -1) 
    { 
      wind_get (w_handle, WF_TOP, &w_handle, 
                &dummy[0], &dummy[1], &dummy[2]); 
      if (get_index_from_handle (w_handle) != -1) 
      { 
        return (TRUE);                      /* it's my own window */ 
      } 
      else                              /* the window is not mine */ 
      { 
        return (FALSE); 
      } 
    } 
  } 
  return (FALSE); 
}                                /* end of function own_top_wind  */ 
 
 
 
/* end of file window.c  */ 
 
